
/* Chrysalide - Outil d'analyse de fichiers binaires
 * logs.c - équivalent Python du fichier "gui/panels/logs.c"
 *
 * Copyright (C) 2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "logs.h"


#include <pygobject.h>


#include <core/logs.h>


#include "../access.h"
#include "../helpers.h"
#include "../pychrysa.h"



/* Fournit la verbosité des messages système. */
static PyObject *py_logs_get_verbosity(PyObject *, PyObject *);

/* Définit la verbosité des messages système. */
static PyObject *py_logs_set_verbosity(PyObject *, PyObject *);

/* Affiche un message dans le journal des messages système. */
static PyObject *py_logs_log_message(PyObject *, PyObject *);

/* Définit les constantes pour les types de message. */
static bool define_python_log_constants(PyObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : self = objet Python concerné par l'appel.                    *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Fournit la verbosité des messages système.                   *
*                                                                             *
*  Retour      : Plus faible niveau des types de message affichés.            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_logs_get_verbosity(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Conversion à retourner      */
    LogMessageType verbosity;               /* Niveau de filtre de message */

    verbosity = get_log_verbosity();

    result = PyLong_FromUnsignedLong(verbosity);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = objet Python concerné par l'appel.                    *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Définit la verbosité des messages système.                   *
*                                                                             *
*  Retour      : Rien en équivalent Python.                                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_logs_set_verbosity(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Bilan à retourner           */
    unsigned long verbosity;                /* Niveau de filtre de message */

    if (!PyArg_ParseTuple(args, "k", &verbosity))
        return NULL;

    set_log_verbosity(verbosity);

    result = Py_None;
    Py_INCREF(result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = objet Python concerné par l'appel.                    *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Affiche un message dans le journal des messages système.     *
*                                                                             *
*  Retour      : Rien en équivalent Python.                                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_logs_log_message(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Bilan à retourner           */
    unsigned long type;                     /* Espèce du message           */
    const char *msg;                        /* Contenu du message          */

    if (!PyArg_ParseTuple(args, "ks", &type, &msg))
        return NULL;

    switch (type)
    {
        case LMT_INFO:
        case LMT_PROCESS:
        case LMT_WARNING:
        case LMT_BAD_BINARY:
        case LMT_ERROR:
        case LMT_EXT_ERROR:
            log_pychrysalide_simple_message(type, msg);
            result = Py_None;
            Py_INCREF(result);
            break;

        default:
            PyErr_SetString(PyExc_ValueError,
                            _("Invalid type of message"));
            result = NULL;
            break;

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : dict = dictionnaire de module à compléter.                   *
*                                                                             *
*  Description : Définit les constantes pour les types de message.            *
*                                                                             *
*  Retour      : true en cas de succès de l'opération, false sinon.           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool define_python_log_constants(PyObject *dict)
{
    bool result;                            /* Bilan à retourner           */

    result = true;

    result &= PyModDict_AddULongMacro(dict, LMT_INFO);
    result &= PyModDict_AddULongMacro(dict, LMT_PROCESS);
    result &= PyModDict_AddULongMacro(dict, LMT_WARNING);
    result &= PyModDict_AddULongMacro(dict, LMT_BAD_BINARY);
    result &= PyModDict_AddULongMacro(dict, LMT_ERROR);
    result &= PyModDict_AddULongMacro(dict, LMT_EXT_ERROR);
    result &= PyModDict_AddULongMacro(dict, LMT_COUNT);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Définit une extension du module 'core' à compléter.          *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool populate_core_module_with_logs(void)
{
    bool result;                            /* Bilan à retourner           */
    PyObject *module;                       /* Module à recompléter        */
    PyObject *dict;                         /* Dictionnaire dudit module   */

    static PyMethodDef py_queue_methods[] = {

        {
            "get_verbosity", py_logs_get_verbosity,
            METH_NOARGS,
            "get_verbosity(, /)\n--\n\nGet the log verbosity."
        },
        {
            "set_verbosity", py_logs_set_verbosity,
            METH_VARARGS,
            "set_verbosity(, /)\n--\n\nSet the log verbosity."
        },
        {
            "log_message", py_logs_log_message,
            METH_VARARGS,
            "log_message(type, msg, /)\n--\n\nDisplay a message in the log window, if any."
        },
        { NULL }

    };

    module = get_access_to_python_module("pychrysalide.core");

    result = register_python_module_methods(module, py_queue_methods);

    if (result)
    {
        dict = PyModule_GetDict(module);
        result = define_python_log_constants(dict);
    }

    return result;

}
